/*******************************************************************************
*
*                   Copyright 2003,MARVELL SEMICONDUCTOR ISRAEL, LTD.
* THIS CODE CONTAINS CONFIDENTIAL INFORMATION OF MARVELL.
* NO RIGHTS ARE GRANTED HEREIN UNDER ANY PATENT, MASK WORK RIGHT OR COPYRIGHT
* OF MARVELL OR ANY THIRD PARTY. MARVELL RESERVES THE RIGHT AT ITS SOLE
* DISCRETION TO REQUEST THAT THIS CODE BE IMMEDIATELY RETURNED TO MARVELL.
* THIS CODE IS PROVIDED "AS IS". MARVELL MAKES NO WARRANTIES, EXPRESSED,
* IMPLIED OR OTHERWISE, REGARDING ITS ACCURACY, COMPLETENESS OR PERFORMANCE.
*
* MARVELL COMPRISES MARVELL TECHNOLOGY GROUP LTD. (MTGL) AND ITS SUBSIDIARIES,
* MARVELL INTERNATIONAL LTD. (MIL), MARVELL TECHNOLOGY, INC. (MTI), MARVELL
* SEMICONDUCTOR, INC. (MSI), MARVELL ASIA PTE LTD. (MAPL), MARVELL JAPAN K.K.
* (MJKK), MARVELL SEMICONDUCTOR ISRAEL. (MSIL),  MARVELL TAIWAN, LTD. AND
* SYSKONNECT GMBH.
*
********************************************************************************
* mvFlash.c - FLASH memory functions and definitions for SV use
*
* DESCRIPTION:                                                               
*       This mvFlash driver gives the user a convenient interface to the FLASH  
*       memory located on the user`s board, it supports various layout
*       configurations such as:
*         1. One pure 8 bit device (Such as AMD`s AM29LV040B).
*         2. 1,2,4 or 8 devices 16 bit wide configured to operate in 8 bit mode.
*         3. 1,2 or 4 devices each 16 bit wide.
*       Before using the driver you must call the initialization function at
*       least once or when ever you are changing the FLASH base address.
*       The list bellow contains the supported FLASH memory devices, new
*       devices can be added easily in the future.
*
* DEPENDENCIES:
*       None.
*
*******************************************************************************/


/*includes*/
#include "mvFlash.h"

/* locals */

/******************************************************************************
* mvFlashParametrs[] , mvFlashTypes[]  
*
*       Those two tables contain the supported mvFlash devices information
*       needed by the driver: 
*       The first table "mvFlashParametrs" starts with 10 shared fields 
*       (currently 6 are reserved):
*       index 0 => Pointer to an entry in the second table list
*       index 1 => baseAddress - Flash memory device base address.
*       index 2 => width - 1, 2, 4 or 8 Bytes. 
*       index 3 => mode - PURE8, X8 or X16 mvFlash configuration
*                 (for X16 devices only)
*       The second table (mvFlashTypes) contains: 
*       Entry`s structure:
*       Manufacture ID,Device ID,number of sectors,list of sector`s sizes
*       (in Kbytes starting with sector number 0).
*       The end of the list is pointed with a zero.
******************************************************************************/
static unsigned long mvFlashParametrs[10]; /* 0  Entry pointer */
                                   /* 1  Base address  */
                                   /* 2  Width         */
                                   /* 3  Mode          */            
                                   /* 4,5,6,7,8,9, spare entries. */
static unsigned long mvFlashTypes[] = {

    /* 0 */          AMD_FLASH,AM29F400BB,11,16,8,8,32,64,64,64,64,64,64,64,
    /* 1 */          AMD_FLASH,AM29F400BT,11,64,64,64,64,64,64,64,32,8,8,16,
    /* 2 */          ST_FLASH,M29W040    ,8,64,64,64,64,64,64,64,64,
    /* 3 */          AMD_FLASH,AM29LV040B,8,64,64,64,64,64,64,64,64,
    /* 4 */          AMD_FLASH,AM29LV800BT,19,64,64,64,64,64,64,64,64,64,64,
                                           64,64,64,64,64,32,8,8,16,
    /* 5 */          INTEL_FLASH,I28F320J3A,32,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,
    /* 6 */          INTEL_FLASH,I28F640J3A,64,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,
    /* 7 */          INTEL_FLASH,I28F128J3A,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
                                     128,128,128,128,128,128,128,128,128,128,
    /* 8 */          AMD_FLASH,AM29LV400BB,11,16,8,8,32,64,64,64,64,64,64,64,
    /* 9 */          AMD_FLASH,AM29LV400BT,11,64,64,64,64,64,64,64,32,8,8,16,
   /* 10 */          INTEL_FLASH,I28F320B3_T,71,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,8,8,8,8,8,8,8,8,
   /* 11 */          INTEL_FLASH,I28F320B3_B,71,8,8,8,8,8,8,8,8,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
   /* 12 */          INTEL_FLASH,I28F160B3_B,39,8,8,8,8,8,8,8,8,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,
   /* 13 */          INTEL_FLASH,I28F160B3_T,39,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,64,64,64,64,64,64,64,64,
                                             64,64,8,8,8,8,8,8,8,8,
   /* 14 */          AMD_FLASH,AM29LV160DT,35,64,64,64,64,64,64,64,64,64,64,64,
                                           64,64,64,64,64,64,64,64,64,64,64,64,
                                           64,64,64,64,64,64,64,64,32,8,8,16,
   /* 15 */          AMD_FLASH,AM29LV160DB,35,16,8,8,32,64,64,64,64,64,64,64,
                                           64,64,64,64,64,64,64,64,64,64,64,64,
                                           64,64,64,64,64,64,64,64,64,64,64,64,
   /* 16 */          AMD_FLASH,AM29LV001BT,10,16,16,16,16,16,16,16,4,4,8,
   /* 17 */          AMD_FLASH,AM29LV001BB,10,8,4,4,16,16,16,16,16,16,16,
                     0 /* End of list indicator */
                    };

/* Local functions */

static unsigned long mvFlashGetSize(void);
static unsigned long mvFlashGetSectorSize(unsigned long);

/* Local Definitions */
#define _1K             0x00000400

/*******************************************************************************
* mvFlashInit - Initializes the FLASH memory driver`s parameters.
*
* DESCRIPTION:
*       This function auto detects flash memory which may be present in
*       the system according to the flashTypes table defined in 
*       gtFlash.c and initializes the global internal data structures 
*       with the flash's base address, width, pointer to flash in the
*       flashTypes table and flash mode (PURE8 , X8 , X16) .If the flash
*       device was not found in the flashTypes table by this function , 
*       the return value will be 0, otherwise the return value will be 
*       the flash's size that was found. this function must be called 
*       at least once before using the flash memory. If the flash's base 
*       address was changed, this function must be called again.
*
* INPUTS:     
*        flashBaseAddress - The physical base address for the flash 
*                          memory device. For example, if the Flash 
*                          Memory is connected to chip select 0 on the 
*                          device interface of the MV device the value
*                          for the base address will be 0x1C00:0000.
*       flashWidth       - The physical width of the flash memory
*                          as connected to the MV's device interface 
*                          (1,2,4 or 8 bytes).
*       flashMode        - Two forms of flash memory are supported by this 
*                          driver. Pure 8 bit devices (PURE8), such as boot 
*                          devices, and 16 bit devices which can be configured 
*                          to operate either as 8 bit devices (X8) or as 16 bit
*                          devices (X16). 
*
* OUTPUT:
*       None.
*
* RETURN:
*       The flash's total size of a successful detection , 0 otherwise .
*
*******************************************************************************/
unsigned long mvFlashInit(unsigned long flashBaseAddress,
                         unsigned long flashWidth, FLASH_DEV_MODE flashMode)
{
    unsigned short mfrId=0;
    unsigned short devId = 0xff;
    unsigned long   FirstAddr, SecondAddr, ThirdAddr;
    unsigned long   pArray=0;
    unsigned long   counter;
    unsigned long   flashSize=0;
    
    /* update the list with relevant parametrs */
    mvFlashParametrs[0] = 0; /* Default initialization */
    mvFlashParametrs[1] = flashBaseAddress;
    mvFlashParametrs[2] = flashWidth;
    mvFlashParametrs[3] = flashMode;
    /* Get the FLASH`s ID */
    switch(FLASH_WIDTH)
    {
        case 1:
            /* AMD or ST ?? **/
            if( flashMode == PURE8 ) /* Boot Flash*/
            {
                FirstAddr = 0x5555;
                SecondAddr= 0x2aaa;
                ThirdAddr = 0x5555;
            }
            else /* X16 device configured to 8bit Mode */
            {
                FirstAddr = 0xaaaa;
                SecondAddr= 0x5555;
                ThirdAddr = 0xaaaa;
            } 
            mvFlashReset();
            MV_WRITE_CHAR(FLASH_BASE_ADDR,  FirstAddr,0xAA);
            MV_WRITE_CHAR(FLASH_BASE_ADDR,  SecondAddr,0x55);
            MV_WRITE_CHAR(FLASH_BASE_ADDR,  ThirdAddr,0x90);
            MV_READ_CHAR(FLASH_BASE_ADDR, 0x0,&mfrId);
            if(mfrId == AMD_FLASH || mfrId == ST_FLASH) 
            {
                mvFlashReset();  
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FirstAddr,0xAA);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  SecondAddr,0x55);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  ThirdAddr,0x90);
                MV_READ_CHAR(FLASH_BASE_ADDR, 0x1,&devId);
                break;
            }
            /* Micron or Intel ?? **/
            MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0xff); /* Read Array */
            /* Flash reset for Intel/Micron */
            MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0x90); /* IDENTIFY Device */   
            MV_READ_CHAR(FLASH_BASE_ADDR, 0x0,&mfrId); 
                                                /*Address for ManufactureID*/
            if(mfrId == INTEL_FLASH || mfrId == MICRON_FLASH)
            {
                MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0xff); /* Read Array */
                /*Flash reset for Intel/Micron*/
                MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0x90); /* IDENTIFY Device */
                MV_READ_CHAR(FLASH_BASE_ADDR, 0x2,&devId); 
                                                /*Address for DeviceID*/
            }

            break; 
        case 2:
        case 4:
        case 8:
            /* AMD or ST ??? */
            mvFlashReset();
            MV_WRITE_SHORT(FLASH_BASE_ADDR, 0x5555 * FLASH_WIDTH,0xaa);
            MV_WRITE_SHORT(FLASH_BASE_ADDR, 0x2aaa * FLASH_WIDTH,0x55);
            MV_WRITE_SHORT(FLASH_BASE_ADDR, 0x5555 * FLASH_WIDTH,0x90);
            MV_READ_SHORT(FLASH_BASE_ADDR,0,&mfrId);
            mvFlashReset();
            /* Read the device ID */
            if(mfrId == AMD_FLASH || mfrId == ST_FLASH) 
            {
                MV_WRITE_SHORT(FLASH_BASE_ADDR, 0x5555 * FLASH_WIDTH,0xaa);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, 0x2aaa * FLASH_WIDTH,0x55);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, 0x5555 * FLASH_WIDTH,0x90);
                MV_READ_SHORT(FLASH_BASE_ADDR,  0x1 * FLASH_WIDTH,&devId);
                break;
            }
            /* Micron or Intel ?? **/
            mvFlashReset();
            MV_WRITE_WORD(FLASH_BASE_ADDR,0,0x00ff00ff); 
            MV_WRITE_WORD(FLASH_BASE_ADDR,0,0x00900090);
            if( (FLASH_WIDTH == 4) || (FLASH_WIDTH == 8) ) /* 32 or 64 bit */
            {
                MV_READ_SHORT(FLASH_BASE_ADDR,0,&mfrId);
            }
            else  /* FLASH_WIDTH = 2 */
            {
                MV_READ_SHORT(FLASH_BASE_ADDR,0,&mfrId);
            }
            if( (mfrId == INTEL_FLASH) || (mfrId == MICRON_FLASH) )
            {
                /* Flash reset for Intel/Micron */
                mvFlashReset();
                MV_WRITE_WORD(FLASH_BASE_ADDR,0,0x00ff00ff); 
                MV_WRITE_WORD(FLASH_BASE_ADDR,0,0x00900090);
                if( (FLASH_WIDTH == 2) ||(FLASH_WIDTH == 4) || 
                    (FLASH_WIDTH == 8) ) /* 32 or 64 bit */
                {
                    MV_READ_SHORT(FLASH_BASE_ADDR, 0x1*FLASH_WIDTH,&devId);
                }
            }
            break;

    }
    /* Try to locate the device in the supported flashes list (FLASE_TYPE). 
       according to the keys: 
       1) mfrId - manufactor ID.
       2) devId - device ID.  
   */
    while(true)
    {

        if (mvFlashTypes[pArray] == 0)
        {
            mvFlashReset();
            return 0;  /* Device not in the list */
        }
        if( (mvFlashTypes[pArray] == mfrId) && 
             (mvFlashTypes[pArray+1] == devId))
        {
            FLASH_POINTER_TO_FLASH = pArray;
            for(counter = 0 ; counter < mvFlashTypes[FLASH_NUMBER_OF_SECTORS] ; 
                counter++)
            {
                flashSize = flashSize + 
                    mvFlashTypes[FLASH_FIRST_SECTOR_SIZE + counter];
            }
            if( FLASH_MODE != PURE8 )
            {
                mvFlashReset();
                return (flashSize * _1K * (FLASH_WIDTH / (FLASH_MODE / 8)) );
            }
            else
            {
                mvFlashReset();
                return (flashSize * _1K * FLASH_WIDTH);
            }
        }
        pArray += (3 + mvFlashTypes[pArray+2]); /* Move to next entry */
    }
}



/*******************************************************************************
* mvFlashEraseSector - Erases a specific sector in the flash memory.
*
* DESCRIPTION:
*       None.
*
* INPUT:
*       sectorNumber - the sector number to be erased.
*
* OUTPUT:
*       None.
*
* RETURN:
*       true on success,false on failure.
*
*******************************************************************************/
bool mvFlashEraseSector(unsigned long sectorNumber)
{ 
    unsigned long    regValue;
    unsigned long    sectorBaseAddress = 0;
    unsigned long    i;
    unsigned long    data20,dataD0,data70;
    unsigned long    dataPoll;
    unsigned long    FirstAddr,SecondAddr,ThirdAddr,FourthAddr,FifthAddr;
    unsigned long    FirstData,SecondData,ThirdData;
    unsigned long    FourthData,FifthData,SixthData;

    /* calculate the sector base Address according to the following parametrs:
       1: FLASH_WIDTH
       2: the size of each sector which it detailed in the table */
    
    /* checking the if the sectorNumber is legal.*/
    if ( sectorNumber > mvFlashTypes[FLASH_NUMBER_OF_SECTORS] - 1) 
        return false;
    /* now the calculation begining of the sector Address */
    for (i=0 ; i < sectorNumber ; i++)
    {
        sectorBaseAddress=sectorBaseAddress + 
            mvFlashTypes[FLASH_FIRST_SECTOR_SIZE+i];    
    }
    /* In case of X8 wide the address should be */
    if( FLASH_MODE == PURE8 )
        sectorBaseAddress = _1K * sectorBaseAddress;
    if( FLASH_MODE == X8 )
        sectorBaseAddress = _1K * sectorBaseAddress;
    /* In case of X16 wide the address should be */
    if( FLASH_MODE == X16 )
        sectorBaseAddress = _1K * sectorBaseAddress / 2;  
    mvFlashReset();
  	if ( (mvFlashTypes[FLASH_POINTER_TO_FLASH] == AMD_FLASH) ||   \
            (mvFlashTypes[FLASH_POINTER_TO_FLASH] == ST_FLASH) )
    {
        switch(FLASH_WIDTH)
        {
            case 1:
                if( FLASH_MODE == PURE8) /* Boot Flash PURE8*/
                {
                    FirstAddr  = 0x5555;
                    SecondAddr = 0x2aaa;
                    ThirdAddr  = 0x5555;
                    FourthAddr = 0x5555; 
                    FifthAddr  = 0x2aaa;
                }
                else
                {
                    FirstAddr  = 0xaaaa;
                    SecondAddr = 0x5555;
                    ThirdAddr  = 0xaaaa;
                    FourthAddr = 0xaaaa; 
                    FifthAddr  = 0x5555;
                }
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FirstAddr,0xAA);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  SecondAddr,0x55);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  ThirdAddr,0x80);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FourthAddr,0xAA);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FifthAddr,0x55);
                MV_WRITE_CHAR(FLASH_BASE_ADDR, ((sectorBaseAddress & 
                                                  0xffffff00)),0x30);
                /* Poll on the flash */
                do{
                    MV_READ_CHAR(FLASH_BASE_ADDR, sectorBaseAddress,
                                 &regValue); 
                } while((regValue & 0x80) != 0x80);

                break;
            case 2:
                if (FLASH_MODE == X16)
                {
                    FirstData  = 0xaa;      /* Data for the First  Cycle    */
                    SecondData = 0x55;      /* Data for the Second Cycle    */
                    ThirdData  = 0x80;      /* Data for the Third  Cycle    */
                    FourthData = 0xaa;      /* Data for the Fourth Cycle    */
                    FifthData  = 0x55;      /* Data for the Fifth  Cycle    */
                    SixthData  = 0x30;      /* Data for the Sixth  Cycle    */
                    FirstAddr  = 0x5555;    /* Address for the First  Cycle */
                    SecondAddr = 0x2aaa;    /* Address for the Second Cycle */
                    ThirdAddr  = 0x5555;    /* Address for the Third  Cycle */
                    FourthAddr = 0x5555;    /* Address for the Fourth Cycle */
                    FifthAddr  = 0x2aaa;    /* Address for the Fifth  Cycle */
                }
                else /* (FLASH_MODE = 8) */
                {
                    FirstData  = 0xaaaa;     /* Data for the First  Cycle    */
                    SecondData = 0x5555;     /* Data for the Second Cycle    */
                    ThirdData  = 0x8080;     /* Data for the Third  Cycle    */
                    FourthData = 0xaaaa;     /* Data for the Fourth Cycle    */
                    FifthData  = 0x5555;     /* Data for the Fifth  Cycle    */
                    SixthData  = 0x3030;     /* Data for the Sixth  Cycle    */
                    FirstAddr  = 0xaaaa;     /* Address for the First  Cycle */
                    SecondAddr = 0x5555;     /* Address for the Second Cycle */
                    ThirdAddr  = 0xaaaa;     /* Address for the Third  Cycle */
                    FourthAddr = 0xaaaa;     /* Address for the Fourth Cycle */
                    FifthAddr  = 0x5555;     /* Address for the Fifth  Cycle */
                }
                MV_WRITE_SHORT(FLASH_BASE_ADDR, FirstAddr * FLASH_WIDTH,
                            FirstData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, SecondAddr * FLASH_WIDTH,
                            SecondData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, ThirdAddr * FLASH_WIDTH,
                            ThirdData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, FourthAddr * FLASH_WIDTH,
                            FourthData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, FifthAddr * FLASH_WIDTH,
                            FifthData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, 
                      (sectorBaseAddress & 0xffffff00)* FLASH_WIDTH,SixthData);
                /* Poll on the flash */
                if (FLASH_MODE == X16) /* 1 device of 16 bit */
                {
                    dataPoll = 0x0080;
                }
                else /* (FLASH_MODE = 8) ==> 2 devices , 8 bit each => 16bit*/
                {
                    dataPoll = 0x8080;
                }
                do{
                    MV_READ_SHORT(FLASH_BASE_ADDR+sectorBaseAddress * 
                                  FLASH_WIDTH,0,&regValue); 
                } while((regValue & dataPoll) != dataPoll);
                break;
            case 4:
                if (FLASH_MODE == X16)
                {
                    FirstData  = 0x00aa00aa; /* Data for the First  Cycle    */
                    SecondData = 0x00550055; /* Data for the Second Cycle    */
                    ThirdData  = 0x00800080; /* Data for the Third  Cycle    */
                    FourthData = 0x00aa00aa; /* Data for the Fourth Cycle    */
                    FifthData  = 0x00550055; /* Data for the Fifth  Cycle    */
                    SixthData  = 0x00300030; /* Data for the Sixth  Cycle    */
                    FirstAddr  = 0x5555;     /* Address for the First  Cycle */
                    SecondAddr = 0x2aaa;     /* Address for the Second Cycle */
                    ThirdAddr  = 0x5555;     /* Address for the Third  Cycle */
                    FourthAddr = 0x5555;     /* Address for the Fourth Cycle */
                    FifthAddr  = 0x2aaa;     /* Address for the Fifth  Cycle */
                }
                else /* if (FLASH_MODE == 8) */
                {
                    FirstData  = 0xaaaaaaaa; /* Data for the First  Cycle    */
                    SecondData = 0x55555555; /* Data for the Second Cycle    */
                    ThirdData  = 0x80808080; /* Data for the Third  Cycle    */
                    FourthData = 0xAAAAAAAA; /* Data for the Fourth Cycle    */
                    FifthData  = 0x55555555; /* Data for the Fifth  Cycle    */
                    SixthData  = 0x30303030; /* Data for the Sixth  Cycle    */
                    FirstAddr  = 0xaaaa;     /* Address for the First  Cycle */
                    SecondAddr = 0x5555;     /* Address for the Second Cycle */
                    ThirdAddr  = 0xaaaa;     /* Address for the Third  Cycle */
                    FourthAddr = 0xaaaa;     /* Address for the Fourth Cycle */
                    FifthAddr  = 0x5555;     /* Address for the Fifth  Cycle */
                }
                MV_WRITE_WORD(FLASH_BASE_ADDR, FirstAddr * FLASH_WIDTH,
                           FirstData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, SecondAddr * FLASH_WIDTH,
                           SecondData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, ThirdAddr * FLASH_WIDTH,
                           ThirdData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, FourthAddr * FLASH_WIDTH,
                           FourthData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, FifthAddr * FLASH_WIDTH,
                           FifthData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, 
                           (sectorBaseAddress & 0xffffff00) * FLASH_WIDTH,
                           SixthData);
                /* Poll on the flash */
                if (FLASH_MODE == X16) /* 4 devices , 16 bit each => 64bit */
                {
                    dataPoll = 0x00800080;
                }
                else /* (FLASH_MODE = 8) ==> 8 devices , 8 bit each => 64bit*/
                {
                    dataPoll = 0x80808080;
                }
                do{
                    MV_READ_WORD(FLASH_BASE_ADDR+sectorBaseAddress * 
                              FLASH_WIDTH,0,&regValue);
                } while((regValue & dataPoll) != dataPoll);
                break;
            case 8: /* In case of 64bit width the transformation is 1->8 */
                if (FLASH_MODE == X16)
                {
                    FirstData  = 0x00aa00aa; /* Data for the First  Cycle    */
                    SecondData = 0x00550055; /* Data for the Second Cycle    */
                    ThirdData  = 0x00800080; /* Data for the Third  Cycle    */
                    FourthData = 0x00aa00aa; /* Data for the Fourth Cycle    */
                    FifthData  = 0x00550055; /* Data for the Fifth  Cycle    */
                    SixthData  = 0x00300030; /* Data for the Sixth  Cycle    */
                    FirstAddr  = 0x5555;     /* Address for the First  Cycle */
                    SecondAddr = 0x2aaa;     /* Address for the Second Cycle */
                    ThirdAddr  = 0x5555;     /* Address for the Third  Cycle */
                    FourthAddr = 0x5555;     /* Address for the Fourth Cycle */
                    FifthAddr  = 0x2aaa;     /* Address for the Fifth  Cycle */
                }
                else /* (FLASH_MODE = 8 */
                {
                    FirstData  = 0xaaaaaaaa; /* Data for the First  Cycle    */
                    SecondData = 0x55555555; /* Data for the Second Cycle    */
                    ThirdData  = 0x80808080; /* Data for the Third  Cycle    */
                    FourthData = 0xAAAAAAAA; /* Data for the Fourth Cycle    */
                    FifthData  = 0x55555555; /* Data for the Fifth  Cycle    */
                    SixthData  = 0x30303030; /* Data for the Sixth  Cycle    */
                    FirstAddr  = 0xaaaa;     /* Address for the First  Cycle */
                    SecondAddr = 0x5555;     /* Address for the Second Cycle */
                    ThirdAddr  = 0xaaaa;     /* Address for the Third  Cycle */
                    FourthAddr = 0xaaaa;     /* Address for the Fourth Cycle */
                    FifthAddr  = 0x5555;     /* Address for the Fifth  Cycle */
                }
                MV_WRITE_WORD(FLASH_BASE_ADDR, FirstAddr * FLASH_WIDTH,
                           FirstData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, SecondAddr * FLASH_WIDTH,
                           SecondData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, ThirdAddr * FLASH_WIDTH,
                           ThirdData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, FourthAddr * FLASH_WIDTH,
                           FourthData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, FifthAddr * FLASH_WIDTH,
                           FifthData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, 
                           (sectorBaseAddress & 0xffffff00) * FLASH_WIDTH,
                           SixthData);
                MV_WRITE_WORD(FLASH_BASE_ADDR,FirstAddr * FLASH_WIDTH + 4,
                           FirstData);
                MV_WRITE_WORD(FLASH_BASE_ADDR,SecondAddr * FLASH_WIDTH + 4,
                           SecondData);
                MV_WRITE_WORD(FLASH_BASE_ADDR,ThirdAddr * FLASH_WIDTH + 4,
                           ThirdData);
                MV_WRITE_WORD(FLASH_BASE_ADDR,FourthAddr * FLASH_WIDTH + 4,
                           FourthData);
                MV_WRITE_WORD(FLASH_BASE_ADDR,FifthAddr * FLASH_WIDTH + 4,
                           FifthData);
                MV_WRITE_WORD(FLASH_BASE_ADDR,(sectorBaseAddress & 0xffffff00) 
                           * FLASH_WIDTH + 4,SixthData);
                /* Poll on the flash */
                if (FLASH_MODE == X16) /* 4 devices , 16 bit each => 64bit */
                {
                    dataPoll = 0x00800080;
                }
                else /* (FLASH_MODE = 8) ==> 8 devices , 8 bit each => 64bit*/
                {
                    dataPoll = 0x80808080;
                }
                do
                {
                    MV_READ_WORD(FLASH_BASE_ADDR,  sectorBaseAddress * 
                                                       FLASH_WIDTH,&regValue);
                } while((regValue & dataPoll) != dataPoll);
                do
                {
                    MV_READ_WORD(FLASH_BASE_ADDR,  sectorBaseAddress * 
                                                   FLASH_WIDTH + 4,&regValue);
                } while((regValue & dataPoll) != dataPoll);
                break;
            default:
                return false;
        }
    } /* End of 'flash erase sector' for AMD/ST */
    else /* Intel/Micron */
    {
        switch(FLASH_WIDTH)
        {
            case 1:
                MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0x20);
                MV_WRITE_CHAR(FLASH_BASE_ADDR, ((sectorBaseAddress & 
                                                  0xffffff00)),0xd0);
                /* Poll on the flash */
                while(true)
                {
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0x70);
                    MV_READ_CHAR(FLASH_BASE_ADDR,0,&regValue);
                    if((regValue & 0x80) == 0x80) 
                        break;
                }
                break;
            case 2:
                if (FLASH_MODE == X16)  /* 1 device 16 bit.  */
                {
                    data20 = 0x0020;;
                    dataD0 = 0x00d0;;
                }
                else /* (FLASH_MODE = 8) ==> 2 devices , 8 bit each => 16bit*/
                {
                    data20 = 0x2020;
                    dataD0 = 0xd0d0;
                }
                MV_WRITE_SHORT(FLASH_BASE_ADDR,0,data20);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, (((sectorBaseAddress*2) & 
                             0xffffff00)),dataD0);
                /* Poll on the flash */
                if (FLASH_MODE == X16)
                {
                    dataPoll = 0x0080;
                    data70 = 0x0070;
                }
                else /* (FLASH_MODE = 8) */
                {
                    dataPoll = 0x8080;
                    data70 = 0x7070;
                }
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR,sectorBaseAddress*2,data70);
                    MV_READ_SHORT(FLASH_BASE_ADDR,
                                  sectorBaseAddress*2,&regValue);
                    if((regValue & 0x0080) == 0x0080) 
                        break;
                }
                break;
            case 4:
                if (FLASH_MODE == X16)  /* 2 devices , 16 bit each => 32bit */
                {
                    data20 = 0x00200020;
                    dataD0 = 0x00d000d0;
                }
                else /* (FLASH_MODE = 8) ==> 4 devices , 8 bit each => 32bit*/
                {
                    data20 = 0x20202020;
                    dataD0 = 0xd0d0d0d0;
                }
                MV_WRITE_WORD(FLASH_BASE_ADDR,0,data20);
                MV_WRITE_WORD(FLASH_BASE_ADDR, (((sectorBaseAddress*4) & 
                                                  0xffffff00)),dataD0);
                /* Poll on the flash */
                if (FLASH_MODE == X16)
                {
                    dataPoll = 0x0080;
                    data70 = 0x0070;
                }
                else /* (FLASH_MODE = 8) */
                {
                    dataPoll = 0x8080;
                    data70 = 0x7070;
                }
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR,0,data70);
                    MV_READ_SHORT(FLASH_BASE_ADDR,0,&regValue);
                    if((regValue & dataPoll) == dataPoll) break;
                }
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR, 2,data70);
                    MV_READ_SHORT(FLASH_BASE_ADDR, 2,&regValue);
                    if((regValue & dataPoll) == dataPoll) break;
                }
                break;
            case 8:
                if (FLASH_MODE == X16)  /* 4 devices , 16 bit each => 64bit */
                {
                    data20 = 0x00200020;
                    dataD0 = 0x00d000d0;
                }
                else /* (FLASH_MODE = 8) ==> 8 devices , 8 bit each => 64bit*/
                {
                    data20 = 0x20202020;
                    dataD0 = 0xd0d0d0d0;
                }
                MV_WRITE_WORD(FLASH_BASE_ADDR,0,data20);
                MV_WRITE_WORD(FLASH_BASE_ADDR, (((sectorBaseAddress*8) & 
                                                  0xffffff00)),dataD0);
                MV_WRITE_WORD(FLASH_BASE_ADDR, 4,data20);
                MV_WRITE_WORD(FLASH_BASE_ADDR, (((sectorBaseAddress*8) & 
                                                  (0xffffff00 + 4))),dataD0);
                /* Poll on the flash */
                if (FLASH_MODE == X16)
                {
                    dataPoll = 0x0080;
                    data70 = 0x0070;
                }
                else /* (FLASH_MODE = 8) */ 
                {
                    dataPoll = 0x8080;
                    data70 = 0x7070;
                }
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR, sectorBaseAddress*8,data70);
                    MV_READ_SHORT(FLASH_BASE_ADDR, sectorBaseAddress*8,
                               &regValue);
                    if((regValue & dataPoll) == dataPoll) break;
                }
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR, 2,data70);
                    MV_READ_SHORT(FLASH_BASE_ADDR, 2,&regValue);
                    if((regValue & dataPoll) == dataPoll) break;
                }
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR, 4,data70);
                    MV_READ_SHORT(FLASH_BASE_ADDR, 4,&regValue);
                    if((regValue & dataPoll) == dataPoll) break;
                }
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR, 6,data70);
                    MV_READ_SHORT(FLASH_BASE_ADDR, 6,&regValue);
                    if((regValue & dataPoll) == dataPoll) break;
                }
                break;
            default:
                return false;
        }
    }
    mvFlashReset();
    return true;
}

/*******************************************************************************
* mvFlashWriteBlock - Write block of chars to the flash.
*
* DESCRIPTION:
*       This function writes a block of data to the Flash memory. Neither the 
*       offset nor the block base address needs to be aligned.
*
* INPUT:
*       offset       - Offset relative to flash memory's base address.
*       numOfByte    - Size of the block in bytes.
*       blockAddress - Pointer to a block of data.
*
* OUTPUT:
*       None.
*
* RETURN:
*       Number of bytes actually written to the flash.
*
*******************************************************************************/
unsigned long mvFlashWriteBlock(unsigned long offset,unsigned long numOfByte,
                               char * blockAddress)
{
    register unsigned long flashWrite;
    register unsigned long align;
    register unsigned long num;
    register unsigned long i;

    if ((offset + numOfByte) >  mvFlashGetSize())
        numOfByte = mvFlashGetSize() - offset; /* getting to flash boundary.*/
    num = numOfByte;
    align = offset % 4;                      /* alignment toward flash.    */
    /* writes chars until the offset toward flash will be align. */
    for(i = align ; (i < 4) && (numOfByte > 0) && (align != 0); i++)
    {
        mvFlashWriteChar(offset,blockAddress[0]);
        numOfByte--;
        offset++;
        blockAddress++;
    }
    while(numOfByte > 3)
    {
        #ifdef LE
        flashWrite = (0xff & blockAddress[0]) | 
            ((0xff & blockAddress[1]) << 8) | 
            ((0xff & blockAddress[2]) << 16) | ((0xff & blockAddress[3]) << 24);
        #else 
        flashWrite = (0xff & blockAddress[3]) | 
            ((0xff & blockAddress[2]) << 8) | 
            ((0xff & blockAddress[1]) << 16) | ((0xff & blockAddress[0]) << 24);
        #endif
        if( flashWrite != 0xffffffff) /* for optimization. */
            mvFlashWriteWord(offset,flashWrite);
        if((offset%0x10000) == 0)
        {
            printf(".");
        }
        numOfByte -= 4;
        blockAddress += 4;
        offset += 4;
    }
    while(numOfByte > 0)
    {
        mvFlashWriteChar(offset,blockAddress[0]);
        numOfByte--;
        blockAddress++;
        offset++;
    }
    return num;
}
/*******************************************************************************
* mvFlashReset - Resets the Flash memory (FLASH`s internal protocol reset).
*
* DESCRIPTION:
*       To perform a write action to a flash device, a certain sequence of 
*       writing (realized using a state machine) operation is needed to be 
*       accomplished  prior to the real data write according to protocol of the
*       flash device manufacture. This function resets ( FLASH`s internal 
*       protocol reset ) the internal state machine incorporated within the 
*       flash needed to perform a write action .
*
* INPUT:
*       Flash memory's internal state machine is reset.
*
* OUTPUT:
*       None.
*
* RETURN:
*       None.
*
*******************************************************************************/
void mvFlashReset(void)
{ 
 	unsigned char   ucData;
    unsigned short  usData;
    unsigned long    uiData;

    if((mvFlashTypes[FLASH_POINTER_TO_FLASH] == AMD_FLASH) || 
                (mvFlashTypes[FLASH_POINTER_TO_FLASH]) == ST_FLASH)
    {
        if (FLASH_MODE == X16)
        {
            ucData = 0xf0;  
            usData = 0xf0;
            uiData = 0x00f000f0;
        }
        else /* case of PURE8 or X8 */
        {
            ucData = 0xf0;  
            usData = 0xf0f0;
            uiData = 0xf0f0f0f0;
        }
    }
    else
    {
        if (FLASH_MODE == X16)
        {
            ucData = 0xff;  
            usData = 0x00ff;
            uiData = 0x00ff00ff;
        }
        else /* case of PURE8 or X8 */
        {
            ucData = 0xff;  
            usData = 0xffff;
            uiData = 0xffffffff;
        }
    }
    switch(FLASH_WIDTH)
    {
        case 1:
            MV_WRITE_CHAR(FLASH_BASE_ADDR,0,ucData);
            break;
        case 2:
            MV_WRITE_SHORT(FLASH_BASE_ADDR,0,usData);
            break;
        case 4:
            MV_WRITE_WORD(FLASH_BASE_ADDR,0,uiData);
            break;
        case 8:
            MV_WRITE_WORD(FLASH_BASE_ADDR,0,uiData);
            MV_WRITE_WORD(FLASH_BASE_ADDR, 0x4,uiData);
            break;
    }
}

                    
/*******************************************************************************
* mvFlashWriteWord - Write 32Bit to the FLASH memory
*
* DESCRIPTION:
*       This function Writes 32 bit data to the flash memory at a given offset 
*       from the FLASH base address in all kind of flash widths. The function 
*       takes care of Big/Little endian conversion. 
*
* INPUT:
*       offset - Offset relative to a flashs base address.
*       data   - 32 bit data to be written to the flash memory.
*
* OUTPUT:
*       None.
*
* RETURN:
*       true on success, false otherwise.
*
*******************************************************************************/
bool mvFlashWriteWord(unsigned long offset,unsigned long data)
{ 
    unsigned char           c,rc;
    unsigned short          s,rs;
    register unsigned long   rw,count=0;
 	register unsigned long   regValue;
    register unsigned long   FirstAddr,SecondAddr,ThirdAddr;
    register unsigned long   FirstData,SecondData,ThirdData;
    register unsigned long   data10,data20,data70,data80,data50;
    if ( (mvFlashTypes[FLASH_POINTER_TO_FLASH] == AMD_FLASH) ||   \
            (mvFlashTypes[FLASH_POINTER_TO_FLASH] == ST_FLASH) )
    {
        switch(FLASH_WIDTH)
        {
            case 1: /* Split the 32 bit write into four 8bit Writings */ 
                                            
                mvFlashReset();
                if( FLASH_MODE == PURE8 ) /* Boot Flash*/
                {
                    FirstAddr  = 0x5555;
                    SecondAddr = 0x2aaa;
                    ThirdAddr  = 0x5555;
                }
                else
                {
                    FirstAddr  = 0xaaaa;
                    SecondAddr = 0x5555;
                    ThirdAddr  = 0xaaaa;
                }
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FirstAddr,0xaa);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  SecondAddr,0x55);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  ThirdAddr,0xa0);
#ifdef LE
                c = data;
#else
                c = (data >> 24);
#endif
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset,c); 
                /* Writing first Byte */
                while(true)
                {
                    MV_READ_CHAR(FLASH_BASE_ADDR, offset,&rc);
                    if( (rc & 0x80) == (c & 0x80) )         /* DQ7 =? DATA */
                        break;                              /* DQ7 =  DATA */
                    if((rc & 0x20) == 0x20)                 /* DQ5 =? '1'  */
                    {
                        MV_READ_CHAR (FLASH_BASE_ADDR, offset,&rc);
                        if((rc & 0x80) == (c & 0x80)) 
                            break;                          /* DQ7 = DATA  */
                        else 

                            return false;                   /* DQ7 != DATA */
                                                }
                }
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FirstAddr,0xaa);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  SecondAddr,0x55);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  ThirdAddr,0xa0);
#ifdef LE
                c = (data >> 8);
#else
                c = (data >> 16);
#endif
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset + 1,c); 
                /* Writing second Byte */
                mvFlashReset();
                while(true)
                {
                    MV_READ_CHAR(FLASH_BASE_ADDR, offset + 1,&rc);
                    if((rc & 0x80) == (c & 0x80))           /* DQ7 =? DATA */
                        break;                              /* DQ7 = DATA  */
                    if((rc & 0x20) == 0x20)                 /* DQ5 =? '1'  */
                    {
                        MV_READ_CHAR(FLASH_BASE_ADDR, offset + 1,&rc);
                        if((rc & 0x80) == (c & 0x80)) 
                            break;                          /* DQ7 = DATA  */
                        else 
                            return false;                   /* DQ7 != DATA */
                    }
                }
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FirstAddr,0xaa);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  SecondAddr,0x55);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  ThirdAddr,0xa0);
#ifdef LE
                c = (data >> 16);
#else
                c = (data >> 8);
#endif
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset + 2,c); 
                /* Writing third Byte */
                mvFlashReset();
                while(true)
                {
                    MV_READ_CHAR(FLASH_BASE_ADDR, offset + 2,&rc);
                    if((rc & 0x80) == (c & 0x80))           /* DQ7 =? DATA */    
                        break;                              /* DQ7 = DATA  */
                    if((rc & 0x20) == 0x20)                 /* DQ5 =? '1'  */
                    {
                        MV_READ_CHAR(FLASH_BASE_ADDR, offset + 2,&rc);
                        if((rc & 0x80) == (c & 0x80)) 
                            break;                          /* DQ7 = DATA  */
                        else                                                 
                            return false;                   /* DQ7 != DATA */
                    }
                }
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  FirstAddr,0xaa);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  SecondAddr,0x55);
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  ThirdAddr,0xa0);
#ifdef LE
                c = (data >> 24);
#else
                c = data;
#endif
                MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset + 3,c); 
                /* Writing fourth Byte */
                mvFlashReset();
                while(true)
                {
                    MV_READ_CHAR(FLASH_BASE_ADDR, offset + 3,&rc);
                    if((rc & 0x80) == (c & 0x80))           /* DQ7 =? DATA */ 
                        break;                              /* DQ7 = DATA  */
                    if((rc & 0x20) == 0x20)                 /* DQ5 =? '1'  */
                    {
                        MV_READ_CHAR(FLASH_BASE_ADDR, offset + 3,&rc);
                        if((rc & 0x80) == (c & 0x80)) 
                            break;                          /* DQ7 = DATA  */
                        else                                                 
                            return false;                   /* DQ7 != DATA */
                    }
                }
                break;
            case 2: /* Split the 32 bit write into two 8/16 bit Writings 
                       (16bit width). */
                if (FLASH_MODE == X16)
                {
                    FirstData  = 0xaa;      /* Data for the First  Cycle    */
                    SecondData = 0x55;      /* Data for the Second Cycle    */
                    ThirdData  = 0xa0;      /* Data for the Third  Cycle    */
                    FirstAddr  = 0x5555;    /* Address for the First  Cycle */
                    SecondAddr = 0x2aaa;    /* Address for the Second Cycle */
                    ThirdAddr  = 0x5555;    /* Address for the Third  Cycle */
                }
                else /* if (FLASH_MODE == 8) */
                {
                    FirstData  = 0xaaaa;     /* Data for the First  Cycle    */
                    SecondData = 0x5555;     /* Data for the Second Cycle    */
                    ThirdData  = 0xa0a0;     /* Data for the Third  Cycle    */
                    FirstAddr  = 0xaaaa;    /* Address for the First  Cycle */
                    SecondAddr = 0x5555;    /* Address for the Second Cycle */
                    ThirdAddr  = 0xaaaa;    /* Address for the Third  Cycle */
                }
                MV_WRITE_SHORT(FLASH_BASE_ADDR, FirstAddr * FLASH_WIDTH,
                                   FirstData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, SecondAddr * FLASH_WIDTH,
                                   SecondData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, ThirdAddr * FLASH_WIDTH,
                                   ThirdData);
#ifdef LE
                s = data;
#else
                s = (data >> 16);
#endif
                MV_WRITE_SHORT(FLASH_BASE_ADDR, offset,s); 
                /* Writing Two Bytes */
                if (FLASH_MODE == X16)
                {
                    data80 = 0x80;;
                    data20 = 0x20;;
                }
                else /* if (FLASH_MODE == 8) */
                {
                    data80 = 0x8080;
                    data20 = 0x2020;
                }
                while(true)
                {
                    MV_READ_SHORT(FLASH_BASE_ADDR, offset,&rs);
                    if((rs & data80) == (s & data80))      /* DQ7 =? DATA */
                        break;                             /* DQ7 =  DATA */
                    if((rs & data20) == data20)            /* DQ5 =? DATA */
                    {
                        MV_READ_SHORT(FLASH_BASE_ADDR, offset,&rs);
                        if((rs & data80) == (s & data80)) 
                            break;                         /* DQ7 = DATA  */
                        else
                        {
                            mvFlashReset();
                            return false;                  /* DQ7 != DATA */
                            
                        }
                    }
                }
                MV_WRITE_SHORT(FLASH_BASE_ADDR, FirstAddr * FLASH_WIDTH,
                                   FirstData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, SecondAddr * FLASH_WIDTH,
                                   SecondData);
                MV_WRITE_SHORT(FLASH_BASE_ADDR, ThirdAddr * FLASH_WIDTH,
                                   ThirdData);
#ifdef LE
                s = (data >> 16);
#else
                s = data;
#endif
                MV_WRITE_SHORT(FLASH_BASE_ADDR, offset + 2,s); 
                /* Writing Two Bytes */
                while(true)
                {
                    MV_READ_SHORT(FLASH_BASE_ADDR, offset + 2,&rs);
                    if((rs & data80) == (s & data80))      /* DQ7 =? DATA */
                        break;                             /* DQ7 =  DATA */
                    if((rs & data20) == data20)            /* DQ5 =? '1'  */
                    {
                        MV_READ_SHORT(FLASH_BASE_ADDR, offset + 2,&rs);
                        if((rs & data80) == (s & data80)) 
                            break;                         /* DQ7 = DATA  */
                        else                                                
                        {
                            mvFlashReset();

                            return false;                  /* DQ7 != DATA */
                        }
                    }
                }
                return true;
            case 4:
            case 8:
                if (FLASH_MODE == X16)
                {
                    FirstData  = 0x00aa00aa;   
                    SecondData = 0x00550055;   
                    ThirdData  = 0x00a000a0;   
                    FirstAddr  = 0x5555; 
                    SecondAddr = 0x2aaa; 
                    ThirdAddr  = 0x5555; 
                }
                else /* (FLASH_MODE == 8) */
                {
                    FirstData  = 0xaaaaaaaa;  /* Data for the First  Cycle    */
                    SecondData = 0x55555555;  /* Data for the Second Cycle    */
                    ThirdData  = 0xa0a0a0a0;  /* Data for the Third  Cycle    */
                    FirstAddr  = 0xaaaaaaaa;  /* Address for the First  Cycle */
                    SecondAddr = 0x55555555;  /* Address for the Second Cycle */
                    ThirdAddr  = 0xaaaaaaaa;  /* Address for the Third  Cycle */
                }
                MV_WRITE_WORD(FLASH_BASE_ADDR, FirstAddr * 
                                 FLASH_WIDTH + offset % FLASH_WIDTH,FirstData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, SecondAddr * 
                                 FLASH_WIDTH + offset % FLASH_WIDTH,SecondData);
                MV_WRITE_WORD(FLASH_BASE_ADDR, ThirdAddr * 
                                 FLASH_WIDTH + offset % FLASH_WIDTH,ThirdData);   
                /* writting the word. */
                MV_WRITE_WORD(FLASH_BASE_ADDR, offset,data);
                /* preparing the polling patterns. */
                if (FLASH_MODE == X16)
                {
                    data80 = 0x00800080;
                    data20 = 0x00200020;
                }
                else /* (FLASH_MODE == 8) */
                {
                    data80 = 0x80808080;
                    data20 = 0x20202020;
                }
                while(true) /* polling loop. */
                {
                    rw = MV_READWORD(FLASH_BASE_ADDR, offset);
                    /* DQ7 =? DATA */
                    if((rw & data80) == (data & data80))     
                        break;                      /* DQ7 =  DATA */
                    if((rw & data20) == data20)     /* DQ5 =? '1'  */
                    {
                        rw = MV_READWORD(FLASH_BASE_ADDR, offset);
                        if((rw & data80) == (data & data80)) 
                            break;                        /* DQ7 = DATA  */
                        else                                               

                            return false;                 /* DQ7 != DATA */
                                            }
                }
                return true;
            default:

                return false; /* case of invalid flash Width. */
        }
    }
    else /* Intel/Micron */
    {
        
        switch(FLASH_WIDTH)
        {
            case 1:
             /* Writing First Byte */
                mvFlashReset();
                count=0;
                while (count<100) 
				{
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0x10);
#ifdef LE                                                                     
                    c = data;                                                 
#else                                                                         
                    c = (data >> 24);                                         
#endif                                                                        
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset,c);                                          
                    while(true)                                               
                    {                                                         
                        /* Reading STATUS Register */                         
                        MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0x70);                  
                        regValue = MV_READCHAR(FLASH_BASE_ADDR,0);
                        if((regValue & 0x80) == 0x80)                         
                            break; /* Case of Write-Operation had Ended */    
                    }                                                         
                    /* Reading STATUS Register for Writing Verification */    
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,0,0x70);              
                    regValue = MV_READCHAR(FLASH_BASE_ADDR,0);
                    if((regValue & 0x10) == 0x10)                             
                        return false; /* Write failure */                     
                    count++;
					mvFlashReset();
                    if (MV_READCHAR(FLASH_BASE_ADDR, offset) == c) 
					{
                        break;
                    }
                }
                /* Writing Second Byte */

#ifdef LE
                c = (data >> 8);
#else
                c = (data >> 16);
#endif
                count = 0;
                while (count<100) {
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  1,0x10);
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset + 1,c);
                    while(true)
                    {
                        /* Reading STATUS Register */
                        MV_WRITE_CHAR(FLASH_BASE_ADDR,  1,0x70);
                        regValue = MV_READCHAR(FLASH_BASE_ADDR, 1);
                        if((regValue & 0x80) == 0x80) 
                            break; /* Write operation ended */
                    }
                    /* Reading STATUS Register for Writing verification */
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  1,0x70); 
                    regValue = MV_READCHAR(FLASH_BASE_ADDR, 1);
                    if((regValue & 0x10) == 0x10) 
                        return false; /* Write failure */

                    count++;
					mvFlashReset();
                    if (MV_READCHAR(FLASH_BASE_ADDR, offset+1) == c) {
                        break;
                    }
                }
                /* Writing Third Byte */
#ifdef LE
                c = (data >> 16);
#else
                c = (data >> 8);
#endif
                count = 0;
                while (count<100) {

                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  2,0x10);
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset + 2,c);
                    while(true)
                    {
                        /* Reading STATUS Register */
                        MV_WRITE_CHAR(FLASH_BASE_ADDR,  2,0x70);
                        regValue = MV_READCHAR(FLASH_BASE_ADDR,  2);
                        if((regValue & 0x80) == 0x80) 
                            break; /* Write operation ended */
                    }
                    /* Reading STATUS Register for Writing Verification */
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  2,0x70); 
                    regValue = MV_READCHAR(FLASH_BASE_ADDR,  2);
                    if((regValue & 0x10) == 0x10) 
                        return false; /* Write failure */


                    count++;
					mvFlashReset();
                    if (MV_READCHAR(FLASH_BASE_ADDR, offset+2) == c) {
                        break;
                    }
                }
                /* Writing Fourth Byte */
#ifdef LE
                c = (data >> 24);
#else
                c = data;
#endif

                count = 0;
                while (count<100) {

                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  3,0x10);
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  offset + 3,c);
                    while(true)
                    {
                        /* Reading STATUS Register */
                        MV_WRITE_CHAR(FLASH_BASE_ADDR,  3,0x70);
                        regValue = MV_READCHAR(FLASH_BASE_ADDR,  3);
                        if((regValue & 0x80) == 0x80) 
                            break; /* Write operation ended */
                    }
                    /* Reading STATUS Register for Writing Verification */
                    MV_WRITE_CHAR(FLASH_BASE_ADDR,  3,0x70); 
                    regValue = MV_READCHAR(FLASH_BASE_ADDR,  3);
                    if((regValue & 0x10) == 0x10) 
                        return false; /* Write failure */


                    count++;
					mvFlashReset();
                    if (MV_READCHAR(FLASH_BASE_ADDR, offset+3) == c) {
                        break;
                    }
                }
                return true;
            case 2:
                if (FLASH_MODE == X16)     /* Case of one X16 bit device */
                {
                    FirstData = 0x0010;    /* Data for the First  Cycle  */
                }
                else /* if (FLASH_MODE == 8) ==> Case of two X8 bit devices */
                {
                    FirstData = 0x1010;    /* Data for the First  Cycle  */
                }
                /* Writing First two Bytes */
                MV_WRITE_SHORT(FLASH_BASE_ADDR,0,FirstData);
#ifdef LE
                s = data;
#else
                s = (data >> 16);
#endif
                MV_WRITE_SHORT(FLASH_BASE_ADDR, offset,s);
                if (FLASH_MODE == X16)
                {
                    data70 = 0x0070;
                    data80 = 0x0080;
                    data10 = 0x0010;
                }
                else /* case of (FLASH_MODE == X8) */
                {
                    data70 = 0x7070;
                    data80 = 0x8080;
                    data10 = 0x1010;
                }
                /* polling on writing action => when done break. */
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR,0,data70);
                    regValue = MV_READSHORT(FLASH_BASE_ADDR,0);
                    if((regValue & data80) == data80) 
                        break;
                }
                /* Reading STATUS Register for Writing Verification */
                MV_WRITE_SHORT(FLASH_BASE_ADDR,0,data70); 
                regValue = MV_READSHORT(FLASH_BASE_ADDR,0);
                if((regValue & data10) == data10) 
                    return false; /* Write failure */
                /* Writing Last two Bytes */
                MV_WRITE_SHORT(FLASH_BASE_ADDR, offset + 2,FirstData);
#ifdef LE
                s = (data >> 16);
#else
                s = data;
#endif
                MV_WRITE_SHORT(FLASH_BASE_ADDR, offset + 2,s);
                /* polling on writing action => when done break. */
                while(true)
                {
                    MV_WRITE_SHORT(FLASH_BASE_ADDR, 2,data70);
                    regValue = MV_READSHORT(FLASH_BASE_ADDR, 2);
                    if((regValue & data80) == data80) 
                        break;
                }
                /* Reading STATUS Register for Writing Verification */
                MV_WRITE_CHAR(FLASH_BASE_ADDR,0,data70); 
                regValue = MV_READCHAR(FLASH_BASE_ADDR,0);
                if((regValue & data10) == data10) 
                    return false; /* Write failure */
                mvFlashReset();
                return true;
            case 4:
            case 8:
                if (FLASH_MODE == X16)      /* Case of one X16 bit device */
                {
                    FirstData = 0x00100010; /* Data for the First  Cycle  */
                }
                else /* (FLASH_MODE == 8) ==> Case of two X8 bit devices */
                {
                    FirstData = 0x10101010; /* Data for the First  Cycle  */
                }
                
                MV_WRITE_WORD(FLASH_BASE_ADDR, offset % FLASH_WIDTH,FirstData);
                
                /* writing the 32-bit data to flash. */
                MV_WRITE_WORD(FLASH_BASE_ADDR, offset,data);
                if (FLASH_MODE == X16)
                {
                    data70 = 0x00700070;
                    data80 = 0x00800080;
                    data10 = 0x00100010;
                    data50 = 0x00500050;
                }
                else /* (FLASH_MODE == 8) */
                {   
                    data70 = 0x70707070;
                    data80 = 0x80808080;
                    data10 = 0x10101010;
                    data50 = 0x50505050;
                }
                /* polling on program completion. */
                while(true)
                {
                    MV_WRITE_WORD(FLASH_BASE_ADDR, 
                                  (offset % FLASH_WIDTH),data70);
                    regValue = MV_READWORD(FLASH_BASE_ADDR,0);
                    if((regValue & data80) == data80)
                        break;
                }
      /* clearing the status word */
                MV_WRITE_WORD(FLASH_BASE_ADDR, (offset % FLASH_WIDTH),data50);
      /* Reading STATUS Register for Writing Verification (programming Error)*/
                MV_WRITE_WORD(FLASH_BASE_ADDR, (offset % FLASH_WIDTH),data70);
                regValue = MV_READWORD(FLASH_BASE_ADDR,
                                       (offset % FLASH_WIDTH));
                if((regValue & data10) == data10)  
                    return false; /* program error */
                mvFlashReset();
                return true;
            default:
                mvFlashReset();
                return false;
        }
    }
    mvFlashReset();
    return true;
}

/*******************************************************************************
* mvFlashReadWord - Read 32Bit from the FLASH memory.
*
* DESCRIPTION:
*       This function read 32 bit data from the flash memory from a given 
*       offset.
*
* INPUT:
*       offset - Offset relative to flash memory's base address.
*
* OUTPUT:
*       None.
*
* RETURN:
*       32 bit data read from the flash memory.
*
*******************************************************************************/
unsigned long mvFlashReadWord(unsigned long offset)
{ 
    return MV_READWORD(FLASH_BASE_ADDR, offset);
}
/*******************************************************************************
* mvFlashReadChar - Read one character form given flash offset.
*
* DESCRIPTION:
*       This function reads 8 bit data from a given offset at the flash memory.
*
* INPUT:
*       offset - Offset relative to flash memory's base address.
*
* OUTPUT:
*       None.
*
* RETURN:
*       8 bit data read from the flash memory.
*
*******************************************************************************/
unsigned char mvFlashReadChar(unsigned long offset)
{ 
    return MV_READCHAR(FLASH_BASE_ADDR,  offset);
}

/*******************************************************************************
* mvFlashReadShort - Read 16bit form given flash offset.
*
* DESCRIPTION:
*       This function reads 16bit data from a given offset at the flash memory.
*
* INPUT:
*       offset - Offset relative to flash memory's base address.
*
* OUTPUT:
*       None.
*
* RETURN:
*       16 bit data read from the flash memory.
*
*******************************************************************************/
unsigned short mvFlashReadShort(unsigned long offset)
{ 
    return MV_READSHORT(FLASH_BASE_ADDR, offset);
}

/*******************************************************************************
* mvFlashWriteShort - write 16bit data to a given flash offset.
*
* DESCRIPTION:
*       Write 16bit data to the flash memory at a given offset. The given offset 
*       must be 2 byte aligned , otherwise the function will return false.
*       The function reads a whole word ( after alignment checking ) and modify 
*       only the part relatively to the endian conversion defined.
*
* INPUT:
*       offset - Offset relative to flash's base address (must be aligned to 2).
*       data   - 16 bit data to be written to the flash memory.
*
* OUTPUT:
*       None.
*
* RETURN:
*       true on success, false otherwise..
*
*******************************************************************************/
bool mvFlashWriteShort(unsigned long offset,unsigned short sdata)
{ 
    unsigned long align;                                                         
    unsigned long flashWrite;                                                         
    unsigned long flashRead;                                                         

    align = offset%4;
    if( (align == 1) || (align == 3) )
        return false;  /* offset misaligned. */
    flashRead = mvFlashReadWord(offset-align);
    if(align == 0)
#ifdef LE
    flashWrite = (flashRead & 0xffff0000) | sdata;
#else
    flashWrite = (flashRead & 0x0000ffff) | (sdata<<16);
#endif
    else /* (align == 2) */
#ifdef LE
    flashWrite = (flashRead & 0x0000ffff) | (sdata<<16);
#else
    flashWrite = (flashRead & 0xffff0000) | sdata;
#endif
    mvFlashWriteWord(offset-align,flashWrite);
    return true;

}

/*******************************************************************************
* mvFlashWriteChar - write one charecter (8 bit) to a given flash offset.
*
* DESCRIPTION:
*       Writes 8 bit data to the flash memory at a given offset.The function 
*       reads a whole word and modify only the part relatively to the endian 
*       conversion defined..
*
* INPUT:
*       offset - Offset relative to flash's base address.
*       data   - 8 bit data to be written to the flash memory.
* 
* OUTPUT:
*       None.
*
* RETURN:
*       None.
*
*******************************************************************************/
bool mvFlashWriteChar(unsigned long offset,char cdata_)
{ 
    unsigned long align;                                                         
    unsigned long flashWrite;                                                         
    unsigned long flashRead;
	unsigned long cdata = cdata_;

	cdata = cdata & 0x000000ff;


    align = offset%4;
    flashRead = mvFlashReadWord(offset-align);
#ifdef LE
	if (align == 0)
	{
		flashWrite = flashRead & 0xffffff00;
		cdata = cdata;
		flashWrite = flashWrite | cdata;
	}
	if (align == 1)
	{
		flashWrite = flashRead & 0xffff00ff;
		cdata = cdata << 8;
		flashWrite = flashWrite | cdata;
	}
	if (align == 2)
	{
		flashWrite = flashRead & 0xff00ffff;
		cdata = cdata << 16;
		flashWrite = flashWrite | cdata;
	}
	if (align == 3)
	{
		flashWrite = flashRead & 0x00ffffff;
		cdata = cdata << 24;
		flashWrite = flashWrite | cdata;
	}
/*    flashWrite = (flashRead & ~(0x000000ff << (8*align)) );  
    flashWrite = flashWrite |  ((cdata) << (8*align) );*/
#else
    flashWrite = (flashRead & ~(0xff000000 >> (8*align)) ) | 
        (cdata << (8*(3-align)) );
#endif

    mvFlashWriteWord(offset-align,flashWrite);
    return true;
}



/*******************************************************************************
* mvFlashGetSize - Return Total flash size.
*
* DESCRIPTION:
*       This function returns the total flash memory's size present on the 
*       system.
*
* INPUT:
*       None.
*
* OUTPUT:
*       None.
*
* RETURN:
*       flash size in case of success , 0 If the flash wasn't initialized.
*
*******************************************************************************/
static unsigned long mvFlashGetSize(void)
{
    unsigned long sectorNum;
    unsigned long totalSize=0;

    if (FLASH_POINTER_TO_FLASH == 0) 
        return 0;                   /* case of flash not initialize */
    for(sectorNum=0 ; sectorNum < mvFlashTypes[FLASH_NUMBER_OF_SECTORS] ; 
        sectorNum++)
    {
        totalSize += mvFlashGetSectorSize(sectorNum);
    }
    return (totalSize);

}

/*******************************************************************************
* mvFlashGetSectorSize - Returns the size of a given sector number.
*
* DESCRIPTION:
*       This function returns the size of a given sector number.
*
* INPUT:
*       sectorNumber - The sector number to get its size.
*
* OUTPUT:
*       None.
*
* RETURN:
*       Sector size. (if Sector number isn't valid or flash wasn't initialize 
*       return 0).
*
*******************************************************************************/
static unsigned long mvFlashGetSectorSize(unsigned long sectorNumber)
{
    if (sectorNumber >= mvFlashTypes[FLASH_NUMBER_OF_SECTORS] ) 
        return 0;
    else
    {
        if( FLASH_MODE != PURE8 )
            return( mvFlashTypes[FLASH_FIRST_SECTOR_SIZE + sectorNumber] * _1K *
                    (FLASH_WIDTH * 8 / FLASH_MODE));
        else /* in case of PUR8 */
            return( mvFlashTypes[FLASH_FIRST_SECTOR_SIZE + sectorNumber] * _1K * 
                    FLASH_WIDTH);
    }
}


/*******************************************************************************
* gtFlashErase - Erases the WHOLE flash memory.
*
* DESCRIPTION:
*       This function erases the WHOLE flash memory device. Use this function as
*       less as possible to avoid shorting the life of the flash device which is
*       erasure limited.
*
* INPUT:
*       None.
*
* OUTPUT:
*       None.
*
* RETURN:
*       true on success,false on failure.
*
*******************************************************************************/
bool mvFlashErase(void)
{ 
  	unsigned long totalFlashSize;
    unsigned long address;
    unsigned long readData;
    unsigned long nextSector;

    mvFlashReset();
    totalFlashSize = mvFlashGetSize();
    /* scan all flash memory space. */
    address = 0;
    while(address < totalFlashSize)
    {
        readData = mvFlashReadWord(address);
        if(readData != 0xffffffff) /* offset with dirty data. */
        {
            mvFlashEraseSector(mvFlashInWhichSector(address));
			printf("Erasing sector %d (%08lx)\n",mvFlashInWhichSector(address),readData);
            nextSector = mvFlashInWhichSector(address) + 1;
            if( nextSector < mvFlashTypes[FLASH_NUMBER_OF_SECTORS])
                /* jump to next sector. */
                address = mvFlashGetSectorOffset(nextSector);
            else
                /* end of erasing. */
                address =  totalFlashSize; 
        }
        else address+=4;
    }
    return true;
}

/*******************************************************************************
* gtFlashInWhichSector - Returns the sector`s number at which offset is at.
*
* DESCRIPTION:
*       This function returns the sector's number in which the offset is 
*       located at.
*
* INPUT:
*       offset - Offset relative to flash memory base address.
*
* OUTPUT:
*       None.
*
* RETURN:
*       Sector number in case of success, 0xFFFFFFFF otherwise.
*
*******************************************************************************/
unsigned long mvFlashInWhichSector(unsigned long offset)
{
    unsigned long sectorNumber,numberOfDevices;
    unsigned long accMemory = 0;

    if( (FLASH_MODE == PURE8) || (FLASH_MODE == X8) )
    {
        numberOfDevices = FLASH_WIDTH;
    }
    else /* X16 mode */
    {
        numberOfDevices = FLASH_WIDTH/2;
    }
    for (sectorNumber=0; sectorNumber < mvFlashTypes[FLASH_NUMBER_OF_SECTORS]; 
                                                               sectorNumber++)
    {
        accMemory = accMemory + 
            mvFlashTypes[FLASH_FIRST_SECTOR_SIZE + sectorNumber]; 
        if(offset < accMemory * numberOfDevices * 1024)    
            return sectorNumber;    
    }
    return 0xffffffff;
}

/*******************************************************************************
* gtFlashGetSectorOffset - Returns sector base address.
*
* DESCRIPTION:
*       This function returns the offset of a given sector relative to the base
*       address of the flash memory.
*
* INPUT:
*       sectorNumber - The sector number to get its size.
*
* OUTPUT:
*       None.
*
* RETURN:
*       Offset of the sector relative to the flash memory base address, or 
*       0xFFFFFFFF in case of failure..
*
*******************************************************************************/
unsigned long mvFlashGetSectorOffset(unsigned long sectorNum)
{
    unsigned long i;
    unsigned long sectorBaseAddress = 0;
    unsigned long numOfDevices;
    
    if( sectorNum > (mvFlashTypes[FLASH_NUMBER_OF_SECTORS]-1) )
        return 0xffffffff;
    for (i=0 ; i < sectorNum ; i++)
    {
        sectorBaseAddress=sectorBaseAddress+
            mvFlashTypes[FLASH_FIRST_SECTOR_SIZE+i];    
    }
    if (FLASH_MODE == X16)
        numOfDevices = FLASH_WIDTH * 8 / FLASH_MODE;
    else
        numOfDevices = FLASH_WIDTH;
    return (_1K * sectorBaseAddress * numOfDevices);

}


